
#include <stdlib.h>
#include <stdio.h>
#include <gtk/gtksignal.h>
#include "gtkmarshalers.h"
#include "gtkcellrendererbutton.h"

static void gtk_cell_renderer_button_get_property  (GObject     *object,
                                                    guint        param_id,
                                                    GValue      *value,
                                                    GParamSpec  *pspec);
static void gtk_cell_renderer_button_set_property  (GObject     *object,
                                                    guint        param_id,
                                                    const GValue *value,
                                                    GParamSpec   *pspec);

static void gtk_cell_renderer_button_init       (GtkCellRendererButton      *celltext);
static void gtk_cell_renderer_button_class_init (GtkCellRendererButtonClass *class);
static void gtk_cell_renderer_button_get_size   (GtkCellRenderer            *cell,
                                                 GtkWidget                  *widget,
                                                 GdkRectangle               *cell_area,
                                                 gint                       *x_offset,
                                                 gint                       *y_offset,
                                                 gint                       *width,
                                                 gint                       *height);
static void gtk_cell_renderer_button_render     (GtkCellRenderer            *cell,
                                                 GdkWindow                  *window,
                                                 GtkWidget                  *widget,
                                                 GdkRectangle               *background_area,
                                                 GdkRectangle               *cell_area,
                                                 GdkRectangle               *expose_area,
                                                 guint                       flags);
static gboolean gtk_cell_renderer_button_activate  (GtkCellRenderer         *cell,
                                                    GdkEvent                *event,
                                                    GtkWidget               *widget,
                                                    const gchar             *path,
                                                    GdkRectangle            *background_area,
                                                    GdkRectangle            *cell_area,
                                                    guint                    flags);


enum {
  CLICKED,
  LAST_SIGNAL
};

enum {
  PROP_ZERO,
  PROP_ACTIVATABLE,
  PROP_BUTTON
};

static guint button_cell_signals[LAST_SIGNAL] = { 0 };

GtkType
gtk_cell_renderer_button_get_type (void)
{
    static GtkType cell_button_type = 0;

    if (!cell_button_type)
    {
        static const GTypeInfo cell_button_info =
        {
            sizeof (GtkCellRendererButtonClass),
            NULL,		/* base_init */
            NULL,		/* base_finalize */
            (GClassInitFunc) gtk_cell_renderer_button_class_init,
            NULL,		/* class_finalize */
            NULL,		/* class_data */
            sizeof (GtkCellRendererButton),
            0,              /* n_preallocs */
            (GInstanceInitFunc) gtk_cell_renderer_button_init,
        };

        cell_button_type = g_type_register_static (GTK_TYPE_CELL_RENDERER, "GtkCellRendererButton", &cell_button_info, 0);
    }

    return cell_button_type;
}

static void
gtk_cell_renderer_button_init (GtkCellRendererButton *cellbutton)
{
    cellbutton->activatable = TRUE;
    cellbutton->button = NULL;
    GTK_CELL_RENDERER (cellbutton)->mode = GTK_CELL_RENDERER_MODE_ACTIVATABLE;
    GTK_CELL_RENDERER (cellbutton)->xpad = 2;
    GTK_CELL_RENDERER (cellbutton)->ypad = 2;
}

static void
gtk_cell_renderer_button_class_init (GtkCellRendererButtonClass *class)
{
    GObjectClass *object_class = G_OBJECT_CLASS (class);
    GtkCellRendererClass *cell_class = GTK_CELL_RENDERER_CLASS (class);

    object_class->get_property = gtk_cell_renderer_button_get_property;
    object_class->set_property = gtk_cell_renderer_button_set_property;

    cell_class->get_size = gtk_cell_renderer_button_get_size;
    cell_class->render = gtk_cell_renderer_button_render;
    cell_class->activate = gtk_cell_renderer_button_activate;

    g_object_class_install_property (object_class,
               PROP_BUTTON,
               g_param_spec_pointer ("button",
                                    "Button widget",
                                    "The widget that contain button",
                                    G_PARAM_READABLE |
                                    G_PARAM_WRITABLE));

    g_object_class_install_property (object_class,
               PROP_ACTIVATABLE,
               g_param_spec_boolean ("activatable",
                                     "Activatable",
                                     "The button button can be activated",
                                     TRUE,
                                     G_PARAM_READABLE |
                                     G_PARAM_WRITABLE));

    button_cell_signals[CLICKED] =
        gtk_signal_new ("clicked",
                        GTK_RUN_LAST,
                        GTK_CLASS_TYPE (object_class),
                        GTK_SIGNAL_OFFSET (GtkCellRendererButtonClass, clicked),
                        _gtk_marshal_VOID__STRING,
                        GTK_TYPE_NONE, 1,
                        GTK_TYPE_STRING);
}

static void
gtk_cell_renderer_button_get_property (GObject     *object,
                                       guint        param_id,
                                       GValue      *value,
                                       GParamSpec  *pspec)
{
    GtkCellRendererButton *cellbutton = GTK_CELL_RENDERER_BUTTON (object);

    switch (param_id)
    {
        case PROP_BUTTON:
            g_value_set_pointer (value, cellbutton->button);
            break;
        case PROP_ACTIVATABLE:
            g_value_set_boolean (value, cellbutton->activatable);
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
            break;
    }
}


static void
gtk_cell_renderer_button_set_property (GObject      *object,
                                       guint         param_id,
                                       const GValue *value,
                                       GParamSpec   *pspec)
{
    GtkCellRendererButton *cellbutton = GTK_CELL_RENDERER_BUTTON (object);

    switch (param_id)
    {
        case PROP_BUTTON:
            cellbutton->button = g_value_get_pointer (value);
            g_object_notify (G_OBJECT(object), "button");
            break;
        case PROP_ACTIVATABLE:
            cellbutton->activatable = g_value_get_boolean (value);
            g_object_notify (G_OBJECT(object), "activatable");
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
            break;
    }
}

GtkCellRenderer *
gtk_cell_renderer_button_new (void)
{
    return GTK_CELL_RENDERER (gtk_type_new (gtk_cell_renderer_button_get_type ()));
}

static void
gtk_cell_renderer_button_get_size (GtkCellRenderer *cell,
                                   GtkWidget       *widget,
                                   GdkRectangle    *cell_area,
                                   gint            *x_offset,
                                   gint            *y_offset,
                                   gint            *width,
                                   gint            *height)
{
    gint calc_width;
    gint calc_height;
    gint w, h;
    GtkCellRendererButton *cellbutton = (GtkCellRendererButton *) cell;
    GtkButton *button = GTK_BUTTON (cellbutton->button);
    if (cell_area)
    {
        w = cell_area->width;
        h = cell_area->height;
    } else
    {
        w = (gint)(GTK_WIDGET(button))->allocation.width;
        h = (gint)(GTK_WIDGET(button))->allocation.height;
    }

    calc_width = cell->xpad * 2 + w;
    calc_height = cell->ypad * 2 + h;

    if (width)
        *width = calc_width;

    if (height)
        *height = calc_height;

    if (cell_area)
    {
        if (x_offset)
        {
            *x_offset = (gint)(cell->xalign * (cell_area->width - calc_width));
            *x_offset = MAX (*x_offset, 0);
        }
        if (y_offset)
        {
            *y_offset = (gint)(cell->yalign * (cell_area->height - calc_height));
            *y_offset = MAX (*y_offset, 0);
        }
    }
}

static void
gtk_cell_renderer_button_render (GtkCellRenderer *cell,
                                 GdkWindow       *window,
                                 GtkWidget       *widget,
                                 GdkRectangle    *background_area,
                                 GdkRectangle    *cell_area,
                                 GdkRectangle    *expose_area,
                                 guint            flags)
{
    GtkCellRendererButton *cellbutton = (GtkCellRendererButton *) cell;
    gint width, height;
    gint x_offset, y_offset;
    GtkShadowType shadow;
    GtkStateType state;

    const gchar *but_label;

    gtk_cell_renderer_button_get_size (cell, widget, cell_area,
                                       &x_offset, &y_offset,
                                       &width, &height);
    width -= cell->xpad*2;
    height -= cell->ypad*2;

    if (width <= 0 || height <= 0)
        return;

    shadow = GTK_BUTTON(cellbutton->button)->depressed ? GTK_SHADOW_IN : GTK_SHADOW_OUT;

    if ((flags & GTK_CELL_RENDERER_SELECTED) == GTK_CELL_RENDERER_SELECTED)
    {
        if (GTK_WIDGET_HAS_FOCUS (widget))
            state = GTK_STATE_SELECTED;
        else
            state = GTK_STATE_ACTIVE;
    }
    if (cellbutton->activatable)
    {
        state = GTK_STATE_NORMAL;
    }
    else
    {
        state = GTK_STATE_INSENSITIVE;
    }
    but_label = gtk_button_get_label (GTK_BUTTON(cellbutton->button));
    printf ("render button %s (%d,%d : %d,%d)\n", but_label,
        x_offset, y_offset, width, height);

    _gtk_button_paint (GTK_BUTTON(cellbutton->button),
                       cell_area,
                       state,
                       shadow,
                       "button",
                       "buttondefault");
}

static gint
gtk_cell_renderer_button_activate (GtkCellRenderer *cell,
                                   GdkEvent        *event,
                                   GtkWidget       *widget,
                                   const gchar     *path,
                                   GdkRectangle    *background_area,
                                   GdkRectangle    *cell_area,
                                   guint            flags)
{
    GtkCellRendererButton *cellbutton;

    cellbutton = GTK_CELL_RENDERER_BUTTON (cell);
    if (cellbutton->activatable)
    {
        gtk_signal_emit (GTK_OBJECT (cell), button_cell_signals[CLICKED], path);
        return TRUE;
    }

    return FALSE;
}

gpointer
gtk_cell_renderer_button_get_button (GtkCellRendererButton *button)
{
    g_return_val_if_fail (GTK_IS_CELL_RENDERER_BUTTON (button), FALSE);

    return button->button;
}

void
gtk_cell_renderer_button_set_button (GtkCellRendererButton *button,
                                     gpointer               widget)
{
  g_return_if_fail (GTK_IS_CELL_RENDERER_BUTTON (button));

  g_object_set (G_OBJECT (button), "button", widget, NULL);
}
